"""
V1 Admin Projects
"""

from ....internals.list_resources import ListResources
from ....internals.post_resources import PostResources


class Projects(ListResources, PostResources):
    """Class for Projects APIs."""

    PROJECTS = '/projects'

    def __init__(self, version, role):
        super().__init__(version, role, Projects.PROJECTS)



"""
    Contains general serializers
"""

import enum
import json
from datetime import date, datetime

from requests import Response


class GeneralObject:
    """Class for General Object"""

    def __init__(self, object_as_dict):
        self.__dict__ = object_as_dict

    def to_json(self):
        """Class method to convert dict to json"""
        return json.loads(serialize(self))

    def get(self, item, default=None):
        """Class method for get item"""
        return self.__dict__.get(item, default)

    def __getattr__(self, item):
        return None

    def __repr__(self):
        return self.__dict__.__str__()


def deserialize(dictionary):
    """General JSON deserializer method"""
    if not isinstance(dictionary, dict):
        return dictionary

    for key, value in dictionary.items():
        if isinstance(value, dict):
            deserialized_value = deserialize(value)
            dictionary[key] = deserialized_value
        elif isinstance(value, list):
            for index, item in enumerate(value):
                deserialized_item = deserialize(item)
                value[index] = deserialized_item
            dictionary[key] = value
    deserialized_object = GeneralObject(dictionary)
    return deserialized_object


class ComplexEncoder(json.JSONEncoder):
    """Class for Custom JSON Encoder"""

    def default(self, o):
        if isinstance(o, Response):
            try:
                encoded_obj = o.json()
            except Exception:
                encoded_obj = {}
        elif isinstance(o, date):
            encoded_obj = o.isoformat()
        elif isinstance(o, datetime):
            encoded_obj = o.isoformat() + "Z"
        elif isinstance(o, enum.Enum):
            encoded_obj = o.value
        elif isinstance(o, GeneralObject):
            encoded_obj = o.__dict__
        elif hasattr(o, '__class__'):
            if hasattr(o, 'serialize'):
                encoded_obj = o.serialize()
            else:
                return None
        elif isinstance(o, Exception):
            encoded_obj = {
                "error": o.__class__.__name__,
                "args": o.args,
            }
        else:
            encoded_obj = json.JSONEncoder.default(self, o)

        return encoded_obj


def serialize(instance, **kwargs):
    """JSON serializer method"""
    return json.dumps(instance, cls=ComplexEncoder, **kwargs)



"""
V1 Admin Invitations
"""
from fyle.platform.internals.post_bulk_resources import PostBulkResources


class Invitations(PostBulkResources):
    """Class for Invitations APIs."""

    INVITATIONS_BULK = '/invitations/bulk'

    def __init__(self, version, role):
        super().__init__(version, role, Invitations.INVITATIONS_BULK)



"""
V1 Admin Categories
"""

from ....internals.list_resources import ListResources
from ....internals.post_resources import PostResources


class Categories(ListResources, PostResources):
    """Class for Categories APIs."""

    CATEGORIES = '/categories'

    def __init__(self, version, role):
        super().__init__(version, role, Categories.CATEGORIES)



"""
    Defines Decorators.
"""

import time

from .auth import Auth
from .. import exceptions as exc


def retry(n, backoff, exceptions):
    """
    Retry Decorator
    Retries the wrapped function/method `times` times if the exceptions listed
    in ``exceptions`` are thrown
    :param n: The number of times to repeat the wrapped function/method
    :param backoff:
    :param exceptions: Lists of exceptions that trigger a retry attempt
    """

    def decorator(func):
        def new_fn(*args, **kwargs):
            attempt = 0
            while attempt < n:
                try:
                    return func(*args, **kwargs)
                except exceptions:
                    if exc.InvalidTokenError in exceptions:
                        Auth().update_access_token()
                    time.sleep(backoff)
                    attempt += 1
            raise exc.RetryException('failed to execute %s despite retrying' % func)

        return new_fn

    return decorator



"""
    Defining Network related operations
"""
# pylint: disable=no-member

import os

import requests

from . import serializers


class Network:
    """Class for making GET, POST requests"""
    HOSTNAME = None
    HEADERS = serializers.deserialize({
        'AUTHORIZATION': 'Authorization',
        'CONTENT_TYPE': 'Content-Type',
        'USER_AGENT': 'User-Agent'
    })

    def __init__(self):
        self.__root_dir = os.path.split(os.getcwd())[1]
        Network.HOSTNAME = os.environ.get('HOSTNAME', self.__root_dir)

    def get_request(self, url, **kwargs):
        """Create a HTTP GET request.

        Parameters:
            url (str): Url for the wanted API.

        Returns:
            A response from the request (dict).
        """

        kwargs.setdefault('allow_redirects', True)
        return self._http_request('GET', url, **kwargs)

    def post_request(self, url, data=None, headers=None, **kwargs):
        """Create a HTTP post request.

        Parameters:
            url (str): Url for the wanted API.
            data:
            headers:
        Returns:
            A response from the request (dict).
        """

        data, headers = Network._process_data_and_headers(data, headers)
        return self._http_request('POST', url, headers=headers, data=data, **kwargs)

    def delete_request(self, url, **kwargs):
        """Create a HTTP delete request.

        Parameters:
            url (str): Url for the wanted API.

        Returns:
            A response from the request (dict).
        """
        return self._http_request('DELETE', url, **kwargs)

    @staticmethod
    def _http_request(method, url, headers=None, **kwargs):
        headers = requests.structures.CaseInsensitiveDict(headers)
        headers[Network.HEADERS.USER_AGENT] = Network.HOSTNAME

        response = requests.request(
            method=method,
            url=url,
            headers=headers,
            **kwargs
        )

        return response

    @classmethod
    def _process_data_and_headers(cls, data, headers):
        headers = requests.structures.CaseInsensitiveDict(headers)

        if isinstance(data, dict):
            data = serializers.serialize(data)
            headers[Network.HEADERS.CONTENT_TYPE] = 'application/json'
        return data, headers



"""
    Defines Delete resource class.
"""

from .api_base import ApiBase


class DeleteResources:
    """Delete Resource Class"""

    def __init__(self, version, role, endpoint):
        self.version = version
        self.role = role
        self.endpoint = endpoint

        self.api = ApiBase(self.version, self.role)

    def delete(self, id_: str):
        """
        Deletes the resource object
        :param id_: resource object id
        :return: Status
        """
        return self.api.make_delete_request(
            api_url='{endpoint}/{id}'.format(endpoint=self.endpoint, id=id_)
        )



"""
    Defines Bulk Post resource class.
"""

from typing import Dict

from .api_base import ApiBase


class PostBulkResources:
    """Bulk Post Resource Class"""

    def __init__(self, version, role, endpoint):
        self.version = version
        self.role = role
        self.endpoint = endpoint
        self.api = ApiBase(self.version, self.role)

    def post_bulk(self, payload: dict) -> Dict:
        """
        Creates or updates resources in bulk.
        :param payload: top-level object containing data which is a array of objects.
            example: {"data": [{object 1}, {object 2}]}
        :return: empty response
        """
        return self.api.make_post_request(
            api_url=self.endpoint,
            payload=payload
        )



"""
    Defines Post resource class.
"""

from typing import Dict

from .api_base import ApiBase


class PostResources:
    """Post Resource Class"""

    def __init__(self, version, role, endpoint):
        self.version = version
        self.role = role
        self.endpoint = endpoint
        self.api = ApiBase(self.version, self.role)

    def post(self, payload: dict) -> Dict:
        """
        Creates or updates expense
        :param payload: Expense object
        :return: expenses Object
        """
        return self.api.make_post_request(
            api_url=self.endpoint,
            payload=payload
        )



"""
V1 Approver Reports
"""

from ....internals.list_resources import ListResources


class Reports(ListResources):
    """Class for Reports APIs."""

    REPORTS = '/reports'

    def __init__(self, version, role):
        super().__init__(version, role, Reports.REPORTS)



"""
V1 Admin Cost Center
"""
from ....internals.list_all_resources import ListAllResources
from ....internals.list_resources import ListResources
from ....internals.post_resources import PostResources


class CostCenters(ListResources, ListAllResources, PostResources):
    """Class for Cost Center APIs."""

    COST_CENTERS = '/cost_centers'

    def __init__(self, version, role):
        super().__init__(version, role, CostCenters.COST_CENTERS)



"""
    Defines List resource class.
"""
from typing import Iterable

from .api_base import ApiBase
from .. import exceptions


class ListAllResources:
    """List All Resource Class"""

    def __init__(self, version, role, endpoint):
        self.version = version
        self.role = role
        self.endpoint = endpoint
        self.api = ApiBase(self.version, self.role)

    def list_all(self, query_params=None) -> Iterable:
        """
        Get Resources
        :param query_params:
        :return: List of Resource Objects
        """
        query_params = {} if query_params is None else query_params
        if 'offset' in query_params or 'limit' in query_params:
            raise exceptions.WrongParamsError('offset, limit should not be used with list_all')
        if 'order' not in query_params:
            raise exceptions.WrongParamsError('Mandatory query params missing: order is mandatory query param.')
        count = 1
        total_count = 0
        while total_count < count:
            resp = self.api.make_get_request(
                api_url=self.endpoint,
                query_params=query_params
            )
            data_size = len(resp.get('data'))
            total_count = total_count + data_size
            count = resp.get('count')
            query_params.update({'offset': total_count, 'limit': data_size})
            yield resp



"""
    Defines Auth Class.
"""

import json

from .network import Network
from .. import exceptions
from ..globals.config import config


class Auth(Network):
    """Auth Class"""

    def __init__(self):
        self.__access_token = None
        super().__init__()

    @staticmethod
    def __get_access_token():
        """
        Get the access token using a HTTP post.

        Returns:
            A new access token.
        """

        api_data = {
            'grant_type': 'refresh_token',
            'refresh_token': config.get('AUTH', 'REFRESH_TOKEN'),
            'client_id': config.get('AUTH', 'CLIENT_ID'),
            'client_secret': config.get('AUTH', 'CLIENT_SECRET')
        }

        token_url = config.get('FYLE', 'TOKEN_URL')
        response = Network().post_request(url=token_url, data=api_data)

        if response.status_code == 200:
            access_token = json.loads(response.text)['access_token']
        elif response.status_code == 401:
            raise exceptions.InvalidTokenError(
                'Wrong client secret or/and refresh token', response.text)
        elif response.status_code == 404:
            raise exceptions.NotFoundItemError('Client ID doesn\'t exist', response.text)
        elif response.status_code == 500:
            raise exceptions.InternalServerError('Internal server error', response.text)
        else:
            raise exceptions.PlatformError(
                'Error: {0}'.format(response.status_code), response.text)

        return access_token

    def update_access_token(self):
        """Update the access token."""

        self.__access_token = self.__get_access_token()
        config.set('AUTH', 'ACCESS_TOKEN', self.__access_token)



"""
    Initializing Fyle Platform SDK
"""

from .exceptions import *
from .platform import Platform

__all__ = [
    'Platform',
    'PlatformError',
    'NotFoundClientError',
    'UnauthorizedClientError',
    'ExpiredTokenError',
    'InvalidTokenError',
    'NoPrivilegeError',
    'WrongParamsError',
    'NotFoundItemError',
    'InternalServerError'
]

name = 'platform'



"""
    Defines List resource class.
"""

from typing import Dict

from fyle.platform import exceptions
from .api_base import ApiBase


class ListResources:
    """List Resource Class"""

    def __init__(self, version, role, endpoint):
        self.version = version
        self.role = role
        self.endpoint = endpoint
        self.api = ApiBase(self.version, self.role)

    def list(self, query_params=None) -> Dict:
        """
        Get Resources
        :param query_params:
        :return: List of Resource Objects
        """
        query_params = {} if query_params is None else query_params
        if not str(query_params.get('offset')) or not str(query_params.get('limit')) or not query_params.get('order'):
            raise exceptions.WrongParamsError(
                'Mandatory query params missing: offset, limit and order are mandatory query params.')
        return self.api.make_get_request(
            api_url=self.endpoint,
            query_params=query_params
        )



"""
    Defines Get resource class.
"""

from typing import Dict

from .api_base import ApiBase


class GetResources:
    """Get Resource Class"""

    def __init__(self, version, role, endpoint):
        self.version = version
        self.role = role
        self.endpoint = endpoint

        self.api = ApiBase(self.version, self.role)

    def get(self, id_: str = None, query_params=None) -> Dict:
        """
        Get Single Resource object by ID
        :param id_: Resource object ID
        :param query_params:
        :return: Resource Object
        """
        query_params = {} if query_params is None else query_params
        if id_:
            api_url = '{endpoint}/{id}'.format(endpoint=self.endpoint, id=id)
        else:
            api_url = self.endpoint

        return self.api.make_get_request(
            api_url=api_url,
            query_params=query_params,
        )



"""
V1 Admin Employees
"""

from ....internals.get_resources import GetResources
from ....internals.list_all_resources import ListAllResources
from ....internals.list_resources import ListResources
from ....internals.post_resources import PostResources


class Employees(ListResources, ListAllResources, PostResources, GetResources):
    """Class for Employees APIs."""

    EMPLOYEES = '/employees'

    def __init__(self, version, role):
        super().__init__(version, role, Employees.EMPLOYEES)



"""
    Define ApiBase class implementing API helper methods.
"""

import json

from .decorators import retry
from .network import Network
from .. import exceptions
from ..globals.config import config


class ApiBase(Network):
    """The base class for all API classes."""

    def __init__(self, version=None, role=None):
        super().__init__()

        self.version = version
        self.role = role

    def _format_api_url(self, endpoint):
        return '{base_url}/{role}{endpoint}'.format(
            base_url=config.get('FYLE', 'SERVER_URL'),
            role=self.role,
            endpoint=endpoint
        )

    @retry(n=3, backoff=5, exceptions=exceptions.InvalidTokenError)
    def make_get_request(self, api_url, query_params=None):
        """Create a HTTP GET request.
        Parameters:
            query_params (dict): HTTP GET parameters for the wanted API.
            api_url (str): Url for the wanted API.
        Returns:
            A response from the request (dict).
        """

        api_headers = {'Authorization': 'Bearer {0}'.format(config.get('AUTH', 'ACCESS_TOKEN'))}
        api_query_params = {}

        for param in query_params:
            # ignore all unused params
            if not query_params[param] is None:
                value = query_params[param]

                # convert boolean to lowercase string
                if isinstance(value, bool):
                    value = str(value).lower()

                api_query_params[param] = value

        response = self.get_request(
            url=self._format_api_url(api_url),
            headers=api_headers,
            params=api_query_params
        )

        if response.status_code == 200:
            result = json.loads(response.text)
            return result

        ApiBase._assert_response(response)

        return None

    @retry(n=3, backoff=5, exceptions=exceptions.InvalidTokenError)
    def make_post_request(self, api_url, payload):
        """Create a HTTP post request.

        Parameters:
            payload (dict): HTTP POST body data for the wanted API.
            api_url (str): Url for the wanted API.

        Returns:
            A response from the request (dict).
        """

        api_headers = {'Authorization': 'Bearer {0}'.format(config.get('AUTH', 'ACCESS_TOKEN')),
                       'Content-Type': 'application/json'}

        response = self.post_request(
            url=self._format_api_url(api_url),
            headers=api_headers,
            data=payload
        )

        if response.status_code == 200:
            result = json.loads(response.text)
            return result

        ApiBase._assert_response(response)

        return None

    @retry(n=3, backoff=5, exceptions=exceptions.InvalidTokenError)
    def make_delete_request(self, api_url):
        """Create a HTTP delete request.

        Parameters:
            api_url (str): Url for the wanted API.

        Returns:
            A response from the request
        """

        api_headers = {'Authorization': 'Bearer {0}'.format(config.get('AUTH', 'ACCESS_TOKEN'))}

        response = self.delete_request(
            url=self._format_api_url(api_url),
            headers=api_headers
        )

        if response.status_code == 204:
            return None

        ApiBase._assert_response(response)

        return None

    @staticmethod
    def _assert_response(response):
        if response.status_code == 400:
            raise exceptions.WrongParamsError(
                'Some of the parameters are wrong', json.loads(response.text))
        if response.status_code == 401:
            raise exceptions.InvalidTokenError(
                'Invalid token, try to refresh it', response.text)
        if response.status_code == 403:
            raise exceptions.NoPrivilegeError(
                'Forbidden, the user has insufficient privilege', response.text)
        if response.status_code == 404:
            raise exceptions.NotFoundItemError(
                'Not found item with ID', response.text)
        if response.status_code == 498:
            raise exceptions.ExpiredTokenError(
                'Expired token, try to refresh it', response.text)
        if response.status_code == 500:
            raise exceptions.InternalServerError(
                'Internal server error', response.text)
        raise exceptions.PlatformError(
            'Error: {0}'.format(response.status_code), response.text)



"""
    Fyle Platform SDK Class
"""

from .apis import v1
from .globals.config import config
from .internals.auth import Auth


class Platform(Auth):
    """The main class which creates a connection with
    Fyle APIs using OAuth2 authentication (refresh token grant type).

    Parameters:
        client_id (str): Client ID for Fyle API.
        client_secret (str): Client secret for Fyle API.
        refresh_token (str): Refresh token for Fyle API.
    """

    def __init__(self, server_url, token_url, client_id, client_secret, refresh_token):
        super().__init__()

        # store the credentials
        self.__server_url = server_url
        self.__token_url = token_url
        self.__client_id = client_id
        self.__client_secret = client_secret
        self.__refresh_token = refresh_token

        self.v1 = v1

        # get the access token
        self.set_server_url()
        self.set_token_url()
        self.set_client_id()
        self.set_client_secret()
        self.set_refresh_token()
        self.update_access_token()

    def set_server_url(self):
        """Set the Server URL in all API objects."""

        config.set('FYLE', 'SERVER_URL', self.__server_url)

    def set_token_url(self):
        """Set the Token URL in all API objects."""

        config.set('FYLE', 'TOKEN_URL', self.__token_url)

    def set_client_id(self):
        """Set the Client ID."""

        config.set('AUTH', 'CLIENT_ID', self.__client_id)

    def set_client_secret(self):
        """Set the Client Secret."""

        config.set('AUTH', 'CLIENT_SECRET', self.__client_secret)

    def set_refresh_token(self):
        """Set the Refresh token."""

        config.set('AUTH', 'REFRESH_TOKEN', self.__refresh_token)



"""
    Initializing Admin API's
"""

from .categories import Categories
from .cost_centers import CostCenters
from .employees import Employees
from .invitations import Invitations
from .projects import Projects
from ..version import version

role = 'admin'

categories = Categories(version, role)
cost_centers = CostCenters(version, role)
employees = Employees(version, role)
projects = Projects(version, role)
invitations = Invitations(version, role)



"""
    Defines Global SDK Configurations
"""

import configparser

configurations = {
    'FYLE': {
        'TOKEN_URL': '',
        'SERVER_URL': ''
    },
    'AUTH': {
        'CLIENT_ID': '',
        'CLIENT_SECRET': '',
        'REFRESH_TOKEN': '',
        'ACCESS_TOKEN': ''
    }
}

config = configparser.ConfigParser()
config.read_dict(configurations)

# from flask import Flask

# class Configurations:
#     """Defines SDK Configurations"""

#     SERVER_URL = None
#     ACCESS_TOKEN = None

# sdk = Flask('fyleplatformsdk')
# sdk.config.from_object(Configurations)
